@@ -19,10 +19,11 @@ from __future__ import print_function
 
 __author__ = 'A. Jesse Jiryu Davis'
 __email__ = 'jesse@mongodb.com'
-__version__ = '1.4.1'
+__version__ = '1.7.0'
 
 import atexit
 import contextlib
+import datetime
 import errno
 import functools
 import inspect
@@ -59,18 +60,23 @@ try:
 except ImportError:
     from cStringIO import StringIO
 
-# Pure-Python bson lib vendored in from PyMongo 3.0.3.
-from mockupdb import _bson
-import mockupdb._bson.codec_options as _codec_options
-import mockupdb._bson.json_util as _json_util
+try:
+    from urllib.parse import quote_plus
+except ImportError:
+    # Python 2
+    from urllib import quote_plus
 
-CODEC_OPTIONS = _codec_options.CodecOptions(document_class=OrderedDict)
+import bson
+from bson import codec_options, json_util
+
+CODEC_OPTIONS = codec_options.CodecOptions(document_class=OrderedDict)
 
 PY3 = sys.version_info[0] == 3
 if PY3:
     string_type = str
     text_type = str
 
+
     def reraise(exctype, value, trace=None):
         raise exctype(str(value)).with_traceback(trace)
 else:
@@ -78,11 +84,10 @@ else:
     text_type = unicode
 
     # "raise x, y, z" raises SyntaxError in Python 3.
-    exec("""def reraise(exctype, value, trace=None):
+    exec ("""def reraise(exctype, value, trace=None):
     raise exctype, str(value), trace
 """)
 
-
 __all__ = [
     'MockupDB', 'go', 'going', 'Future', 'wait_until', 'interactive_server',
 
@@ -121,7 +126,7 @@ def go(fn, *args, **kwargs):
     'return value'
     """
     if not callable(fn):
-        raise TypeError('go() requires a function, not %r' % (fn, ))
+        raise TypeError('go() requires a function, not %r' % (fn,))
     result = [None]
     error = []
 
@@ -284,49 +289,6 @@ def _get_c_string(data, position):
     return _utf_8_decode(data[position:end], None, True)[0], end + 1
 
 
-def _bson_values_equal(a, b):
-    # Check if values are either from our vendored bson or PyMongo's bson.
-    for value in (a, b):
-        if not hasattr(value, '_type_marker'):
-            # Normal equality.
-            return a == b
-
-    def marker(obj):
-        return getattr(obj, '_type_marker', None)
-
-    if marker(a) != marker(b):
-        return a == b
-
-    # Instances of Binary, ObjectId, etc. from our vendored bson don't equal
-    # instances of PyMongo's bson classes that users pass in as message specs,
-    # since isinstance() fails. Reimplement equality checks for each class.
-    key_fn = {
-        # Binary.
-        5: lambda obj: (obj.subtype, bytes(obj)),
-        # ObjectId.
-        7: lambda obj: obj.binary,
-        # Regex.
-        11: lambda obj: (obj.pattern, obj.flags),
-        # Code.
-        13: lambda obj: (obj.scope, str(obj)),
-        # Timestamp.
-        17: lambda obj: (obj.time, obj.inc),
-        # Decimal128.
-        19: lambda obj: obj.bid,
-        # DBRef.
-        100: lambda obj: (obj.database, obj.collection, obj.id),
-        # MaxKey.
-        127: lambda obj: 127,
-        # MinKey.
-        255: lambda obj: 255,
-    }.get(marker(a))
-
-    if key_fn:
-        return key_fn(a) == key_fn(b)
-
-    return a == b
-
-
 class _PeekableQueue(Queue):
     """Only safe from one consumer thread at a time."""
     _NO_ITEM = object()
@@ -351,6 +313,19 @@ class _PeekableQueue(Queue):
             return Queue.get(self, block, timeout)
 
 
+def _ismap(obj):
+    return isinstance(obj, Mapping)
+
+
+def _islist(obj):
+    return isinstance(obj, list)
+
+
+def _dt_rounded(dt):
+    """Python datetimes have microsecond precision, BSON only milliseconds."""
+    return dt.replace(microsecond=dt.microsecond - dt.microsecond % 1000)
+
+
 class Request(object):
     """Base class for `Command`, `OpMsg`, and so on.
 
@@ -387,7 +362,7 @@ class Request(object):
         self._verbose = self._server and self._server.verbose
         self._server_port = kwargs.pop('server_port', None)
         self._docs = make_docs(*args, **kwargs)
-        if not all(isinstance(doc, Mapping) for doc in self._docs):
+        if not all(_ismap(doc) for doc in self._docs):
             raise_args_err()
 
     @property
@@ -432,7 +407,12 @@ class Request(object):
     @property
     def client_port(self):
         """Client connection's TCP port."""
-        return self._client.getpeername()[1]
+        address = self._client.getpeername()
+        if isinstance(address, tuple):
+            return address[1]
+
+        # Maybe a Unix domain socket connection.
+        return 0
 
     @property
     def server(self):
@@ -505,23 +485,56 @@ class Request(object):
 
     def _matches_docs(self, docs, other_docs):
         """Overridable method."""
-        for i, doc in enumerate(docs):
-            other_doc = other_docs[i]
-            for key, value in doc.items():
-                if value is absent:
-                    if key in other_doc:
-                        return False
-                elif not _bson_values_equal(value, other_doc.get(key, None)):
-                    return False
-            if isinstance(doc, (OrderedDict, _bson.SON)):
-                if not isinstance(other_doc, (OrderedDict, _bson.SON)):
-                    raise TypeError(
-                        "Can't compare ordered and unordered document types:"
-                        " %r, %r" % (doc, other_doc))
-                keys = [key for key, value in doc.items()
-                        if value is not absent]
-                if not seq_match(keys, list(other_doc.keys())):
+        for doc, other_doc in zip(docs, other_docs):
+            if not self._match_map(doc, other_doc):
+                return False
+
+        return True
+
+    def _match_map(self, doc, other_doc):
+        for key, val in doc.items():
+            if val is absent:
+                if key in other_doc:
                     return False
+            elif not self._match_val(val, other_doc.get(key, None)):
+                return False
+
+        if isinstance(doc, (OrderedDict, bson.SON)):
+            if not isinstance(other_doc, (OrderedDict, bson.SON)):
+                raise TypeError(
+                    "Can't compare ordered and unordered document types:"
+                    " %r, %r" % (doc, other_doc))
+            keys = [key for key, val in doc.items()
+                    if val is not absent]
+            if not seq_match(keys, list(other_doc.keys())):
+                return False
+
+        return True
+
+    def _match_list(self, lst, other_lst):
+        if len(lst) != len(other_lst):
+            return False
+
+        for val, other_val in zip(lst, other_lst):
+            if not self._match_val(val, other_val):
+                return False
+
+        return True
+
+    def _match_val(self, val, other_val):
+        if _ismap(val) and _ismap(other_val):
+            if not self._match_map(val, other_val):
+                return False
+        elif _islist(val) and _islist(other_val):
+            if not self._match_list(val, other_val):
+                return False
+        elif (isinstance(val, datetime.datetime)
+              and isinstance(other_val, datetime.datetime)):
+            if _dt_rounded(val) != _dt_rounded(other_val):
+                return False
+        elif val != other_val:
+            return False
+
         return True
 
     def _replies(self, *args, **kwargs):
@@ -570,7 +583,7 @@ class CommandBase(Request):
     is_command = True
 
     # Check command name case-insensitively.
-    _non_matched_attrs = Request._non_matched_attrs + ('command_name', )
+    _non_matched_attrs = Request._non_matched_attrs + ('command_name',)
 
     @property
     def command_name(self):
@@ -595,7 +608,7 @@ class CommandBase(Request):
         if items and other_items:
             if items[0][0].lower() != other_items[0][0].lower():
                 return False
-            if not _bson_values_equal(items[0][1], other_items[0][1]):
+            if items[0][1] != other_items[0][1]:
                 return False
         return super(CommandBase, self)._matches_docs(
             [OrderedDict(items[1:])],
@@ -617,9 +630,9 @@ class OpMsg(CommandBase):
         """
         flags, = _UNPACK_UINT(msg[:4])
         pos = 4
-        first_payload_type, = _UNPACK_BYTE(msg[pos:pos+1])
+        first_payload_type, = _UNPACK_BYTE(msg[pos:pos + 1])
         pos += 1
-        first_payload_size, = _UNPACK_INT(msg[pos:pos+4])
+        first_payload_size, = _UNPACK_INT(msg[pos:pos + 4])
         if flags != 0 and flags != 2:
             raise ValueError('OP_MSG flag must be 0 or 2 not %r' % (flags,))
         if first_payload_type != 0:
@@ -627,21 +640,21 @@ class OpMsg(CommandBase):
                 first_payload_type,))
 
         # Parse the initial document and add the optional payload type 1.
-        payload_document = _bson.decode_all(msg[pos:pos+first_payload_size],
-                                            CODEC_OPTIONS)[0]
+        payload_document = bson.decode_all(msg[pos:pos + first_payload_size],
+                                           CODEC_OPTIONS)[0]
         pos += first_payload_size
         if len(msg) != pos:
-            payload_type, = _UNPACK_BYTE(msg[pos:pos+1])
+            payload_type, = _UNPACK_BYTE(msg[pos:pos + 1])
             pos += 1
             if payload_type != 1:
                 raise ValueError('Second OP_MSG payload type must be 1 not %r'
                                  % (payload_type,))
-            section_size, = _UNPACK_INT(msg[pos:pos+4])
+            section_size, = _UNPACK_INT(msg[pos:pos + 4])
             if len(msg) != pos + section_size:
                 raise ValueError('More than two OP_MSG sections unsupported')
             pos += 4
             identifier, pos = _get_c_string(msg, pos)
-            documents = _bson.decode_all(msg[pos:], CODEC_OPTIONS)
+            documents = bson.decode_all(msg[pos:], CODEC_OPTIONS)
             payload_document[identifier] = documents
 
         database = payload_document['$db']
@@ -684,7 +697,7 @@ class OpMsg(CommandBase):
         else:
             if len(reply.docs) > 1:
                 raise ValueError('OP_MSG reply with multiple documents: %s'
-                                 % (reply.docs, ))
+                                 % (reply.docs,))
             reply.doc.setdefault('ok', 1)
         super(OpMsg, self)._replies(reply)
 
@@ -713,7 +726,7 @@ class OpQuery(Request):
         pos += 4
         num_to_return, = _UNPACK_INT(msg[pos:pos + 4])
         pos += 4
-        docs = _bson.decode_all(msg[pos:], CODEC_OPTIONS)
+        docs = bson.decode_all(msg[pos:], CODEC_OPTIONS)
         if is_command:
             assert len(docs) == 1
             command_ns = namespace[:-len('.$cmd')]
@@ -733,7 +746,7 @@ class OpQuery(Request):
 
     def __init__(self, *args, **kwargs):
         fields = kwargs.pop('fields', None)
-        if fields is not None and not isinstance(fields, Mapping):
+        if fields is not None and not _ismap(fields):
             raise_args_err()
         self._fields = fields
         self._num_to_skip = kwargs.pop('num_to_skip', None)
@@ -780,7 +793,7 @@ class Command(CommandBase, OpQuery):
         else:
             if len(reply.docs) > 1:
                 raise ValueError('Command reply with multiple documents: %s'
-                                 % (reply.docs, ))
+                                 % (reply.docs,))
             reply.doc.setdefault('ok', 1)
         super(Command, self)._replies(reply)
 
@@ -798,6 +811,7 @@ class Command(CommandBase, OpQuery):
 
 class OpGetMore(Request):
     """An OP_GET_MORE the client executes on the server."""
+
     @classmethod
     def unpack(cls, msg, client, server, request_id):
         """Parse message and return an `OpGetMore`.
@@ -832,6 +846,7 @@ class OpGetMore(Request):
 
 class OpKillCursors(Request):
     """An OP_KILL_CURSORS the client executes on the server."""
+
     @classmethod
     def unpack(cls, msg, client, server, _):
         """Parse message and return an `OpKillCursors`.
@@ -844,7 +859,7 @@ class OpKillCursors(Request):
         cursor_ids = []
         pos = 8
         for _ in range(num_of_cursor_ids):
-            cursor_ids.append(_UNPACK_INT(msg[pos:pos+4])[0])
+            cursor_ids.append(_UNPACK_INT(msg[pos:pos + 4])[0])
             pos += 4
         return OpKillCursors(_client=client, cursor_ids=cursor_ids,
                              _server=server)
@@ -880,7 +895,7 @@ class OpInsert(_LegacyWrite):
         """
         flags, = _UNPACK_INT(msg[:4])
         namespace, pos = _get_c_string(msg, 4)
-        docs = _bson.decode_all(msg[pos:], CODEC_OPTIONS)
+        docs = bson.decode_all(msg[pos:], CODEC_OPTIONS)
         return cls(*docs, namespace=namespace, flags=flags, _client=client,
                    request_id=request_id, _server=server)
 
@@ -900,7 +915,7 @@ class OpUpdate(_LegacyWrite):
         # First 4 bytes of OP_UPDATE are "reserved".
         namespace, pos = _get_c_string(msg, 4)
         flags, = _UNPACK_INT(msg[pos:pos + 4])
-        docs = _bson.decode_all(msg[pos + 4:], CODEC_OPTIONS)
+        docs = bson.decode_all(msg[pos + 4:], CODEC_OPTIONS)
         return cls(*docs, namespace=namespace, flags=flags, _client=client,
                    request_id=request_id, _server=server)
 
@@ -920,13 +935,14 @@ class OpDelete(_LegacyWrite):
         # First 4 bytes of OP_DELETE are "reserved".
         namespace, pos = _get_c_string(msg, 4)
         flags, = _UNPACK_INT(msg[pos:pos + 4])
-        docs = _bson.decode_all(msg[pos + 4:], CODEC_OPTIONS)
+        docs = bson.decode_all(msg[pos + 4:], CODEC_OPTIONS)
         return cls(*docs, namespace=namespace, flags=flags, _client=client,
                    request_id=request_id, _server=server)
 
 
 class Reply(object):
     """A reply from `MockupDB` to the client."""
+
     def __init__(self, *args, **kwargs):
         self._flags = kwargs.pop('flags', 0)
         self._docs = make_docs(*args, **kwargs)
@@ -956,6 +972,7 @@ class Reply(object):
 
 class OpReply(Reply):
     """An OP_REPLY reply from `MockupDB` to the client."""
+
     def __init__(self, *args, **kwargs):
         self._cursor_id = kwargs.pop('cursor_id', 0)
         self._starting_from = kwargs.pop('starting_from', 0)
@@ -991,7 +1008,7 @@ class OpReply(Reply):
         response_to = request.request_id
 
         data = b''.join([flags, cursor_id, starting_from, number_returned])
-        data += b''.join([_bson.BSON.encode(doc) for doc in self._docs])
+        data += b''.join([bson.BSON.encode(doc) for doc in self._docs])
 
         message = struct.pack("<i", 16 + len(data))
         message += struct.pack("<i", reply_id)
@@ -1002,6 +1019,7 @@ class OpReply(Reply):
 
 class OpMsgReply(Reply):
     """A OP_MSG reply from `MockupDB` to the client."""
+
     def __init__(self, *args, **kwargs):
         super(OpMsgReply, self).__init__(*args, **kwargs)
         assert len(self._docs) <= 1, 'OpMsgReply can only have one document'
@@ -1031,7 +1049,7 @@ class OpMsgReply(Reply):
         """Take a `Request` and return an OP_MSG message as bytes."""
         flags = struct.pack("<I", self._flags)
         payload_type = struct.pack("<b", 0)
-        payload_data = _bson.BSON.encode(self.doc)
+        payload_data = bson.BSON.encode(self.doc)
         data = b''.join([flags, payload_type, payload_data])
 
         reply_id = random.randint(0, 1000000)
@@ -1063,6 +1081,7 @@ class Matcher(object):
     and by `~MockupDB.got` to test if it did and return ``True`` or ``False``.
     Used by `.autoresponds` to match requests with autoresponses.
     """
+
     def __init__(self, *args, **kwargs):
         self._kwargs = kwargs
         self._prototype = make_prototype_request(*args, **kwargs)
@@ -1103,6 +1122,7 @@ class Matcher(object):
 
 def _synchronized(meth):
     """Call method while holding a lock."""
+
     @functools.wraps(meth)
     def wrapper(self, *args, **kwargs):
         with self._lock:
@@ -1145,7 +1165,7 @@ class _AutoResponder(object):
                 # ourselves in __init__.
                 request.replies(*self._args, **self._kwargs)
                 return True
-            
+
     def cancel(self):
         """Stop autoresponding."""
         self._server.cancel_responder(self)
@@ -1194,11 +1214,25 @@ class MockupDB(object):
         if `auto_ismaster` is True, default 0.
       - `max_wire_version`: the maxWireVersion to include in ismaster responses
         if `auto_ismaster` is True, default 6.
+      - `uds_path`: a Unix domain socket path. MockupDB will attempt to delete
+        the path if it already exists.
     """
+
     def __init__(self, port=None, verbose=False,
                  request_timeout=10, auto_ismaster=None,
-                 ssl=False, min_wire_version=0, max_wire_version=6):
-        self._address = ('localhost', port)
+                 ssl=False, min_wire_version=0, max_wire_version=6,
+                 uds_path=None):
+        if port is not None and uds_path is not None:
+            raise TypeError(
+                ("You can't pass port=%s and uds_path=%s,"
+                 " pass only one or neither") % (port, uds_path))
+
+        self._uds_path = uds_path
+        if uds_path:
+            self._address = (uds_path, 0)
+        else:
+            self._address = ('localhost', port)
+
         self._verbose = verbose
         self._label = None
         self._ssl = ssl
@@ -1231,8 +1265,12 @@ class MockupDB(object):
 
     @_synchronized
     def run(self):
-        """Begin serving. Returns the bound port."""
-        self._listening_sock, self._address = bind_socket(self._address)
+        """Begin serving. Returns the bound port, or 0 for domain socket."""
+        self._listening_sock, self._address = (
+            bind_domain_socket(self._address)
+            if self._uds_path
+            else bind_tcp_socket(self._address))
+
         if self._ssl:
             certfile = os.path.join(os.path.dirname(__file__), 'server.pem')
             self._listening_sock = _ssl.wrap_socket(
@@ -1266,6 +1304,12 @@ class MockupDB(object):
             for thread in threads:
                 thread.join(10)
 
+        if self._uds_path:
+            try:
+                os.unlink(self._uds_path)
+            except OSError:
+                pass
+
     def receives(self, *args, **kwargs):
         """Pop the next `Request` and assert it matches.
 
@@ -1467,7 +1511,7 @@ class MockupDB(object):
 
     subscribe = autoresponds
     """Synonym for `.autoresponds`."""
-    
+
     @_synchronized
     def cancel_responder(self, responder):
         """Cancel a responder that was registered with `autoresponds`."""
@@ -1481,7 +1525,7 @@ class MockupDB(object):
     @property
     def address_string(self):
         """The listening "host:port"."""
-        return '%s:%d' % self._address
+        return format_addr(self._address)
 
     @property
     def host(self):
@@ -1496,8 +1540,10 @@ class MockupDB(object):
     @property
     def uri(self):
         """Connection string to pass to `~pymongo.mongo_client.MongoClient`."""
-        assert self.host and self.port
-        uri = 'mongodb://%s:%s' % self._address
+        if self._uds_path:
+            uri = 'mongodb://%s' % (quote_plus(self._uds_path),)
+        else:
+            uri = 'mongodb://%s' % (format_addr(self._address),)
         return uri + '/?ssl=true' if self._ssl else uri
 
     @property
@@ -1555,7 +1601,7 @@ class MockupDB(object):
                 if select.select([self._listening_sock.fileno()], [], [], 1):
                     client, client_addr = self._listening_sock.accept()
                     client.setblocking(True)
-                    self._log('connection from %s:%s' % client_addr)
+                    self._log('connection from %s' % format_addr(client_addr))
                     server_thread = threading.Thread(
                         target=functools.partial(
                             self._server_loop, client, client_addr))
@@ -1569,7 +1615,7 @@ class MockupDB(object):
                     server_thread.start()
             except socket.error as error:
                 if error.errno not in (
-                        errno.EAGAIN, errno.EBADF, errno.EWOULDBLOCK):
+                    errno.EAGAIN, errno.EBADF, errno.EWOULDBLOCK):
                     raise
             except select.error as error:
                 if error.args[0] == errno.EBADF:
@@ -1611,7 +1657,7 @@ class MockupDB(object):
                 traceback.print_exc()
                 break
 
-        self._log('disconnected: %s:%d' % client_addr)
+        self._log('disconnected: %s' % format_addr(client_addr))
         client.close()
 
     def _log(self, msg):
@@ -1642,10 +1688,24 @@ class MockupDB(object):
     __next__ = next
 
     def __repr__(self):
+        if self._uds_path:
+            return 'MockupDB(uds_path=%s)' % (self._uds_path,)
+
         return 'MockupDB(%s, %s)' % self._address
 
 
-def bind_socket(address):
+def format_addr(address):
+    """Turn a TCP or Unix domain socket address into a string."""
+    if isinstance(address, tuple):
+        if address[1]:
+            return '%s:%d' % address
+        else:
+            return address[0]
+
+    return address
+
+
+def bind_tcp_socket(address):
     """Takes (host, port) and returns (socket_object, (host, port)).
 
     If the passed-in port is None, bind an unused port and return it.
@@ -1669,6 +1729,20 @@ def bind_socket(address):
     raise socket.error('could not bind socket')
 
 
+def bind_domain_socket(address):
+    """Takes (socket path, 0) and returns (socket_object, (path, 0))."""
+    path, _ = address
+    try:
+        os.unlink(path)
+    except OSError:
+        pass
+
+    sock = socket.socket(socket.AF_UNIX)
+    sock.bind(path)
+    sock.listen(128)
+    return sock, (path, 0)
+
+
 OPCODES = {OP_MSG: OpMsg,
            OP_QUERY: OpQuery,
            OP_INSERT: OpInsert,
@@ -1849,7 +1923,7 @@ def docs_repr(*args):
     >>> print(docs_repr(OrderedDict([(u'ts', now)])))
     {"ts": {"$date": 123456000}}
     >>>
-    >>> oid = _bson.ObjectId(b'123456781234567812345678')
+    >>> oid = bson.ObjectId(b'123456781234567812345678')
     >>> print(docs_repr(OrderedDict([(u'oid', oid)])))
     {"oid": {"$oid": "123456781234567812345678"}}
     """
@@ -1857,7 +1931,7 @@ def docs_repr(*args):
     for doc_idx, doc in enumerate(args):
         if doc_idx > 0:
             sio.write(u', ')
-        sio.write(text_type(_json_util.dumps(doc)))
+        sio.write(text_type(json_util.dumps(doc)))
     return sio.getvalue()
 
 
@@ -1923,7 +1997,7 @@ def raise_args_err(message='bad arguments', error_class=TypeError):
 
 
 def interactive_server(port=27017, verbose=True, all_ok=False, name='MockupDB',
-                       ssl=False):
+                       ssl=False, uds_path=None):
     """A `MockupDB` that the mongo shell can connect to.
 
     Call `~.MockupDB.run` on the returned server, and clean it up with
@@ -1932,11 +2006,15 @@ def interactive_server(port=27017, verbose=True, all_ok=False, name='MockupDB',
     If ``all_ok`` is True, replies {ok: 1} to anything unmatched by a specific
     responder.
     """
+    if uds_path is not None:
+        port = None
+
     server = MockupDB(port=port,
                       verbose=verbose,
                       request_timeout=int(1e6),
                       ssl=ssl,
-                      auto_ismaster=True)
+                      auto_ismaster=True,
+                      uds_path=uds_path)
     if all_ok:
         server.autoresponds({})
     server.autoresponds('whatsmyuri', you='localhost:12345')
@@ -1945,4 +2023,5 @@ def interactive_server(port=27017, verbose=True, all_ok=False, name='MockupDB',
     server.autoresponds(OpMsg('buildInfo'), version='MockupDB ' + __version__)
     server.autoresponds(OpMsg('listCollections'))
     server.autoresponds('replSetGetStatus', ok=0)
+    server.autoresponds('getFreeMonitoringStatus', ok=0)
     return server
