@@ -4,8 +4,11 @@
 """Test MockupDB."""
 
 import contextlib
+import datetime
+import os
 import ssl
 import sys
+import tempfile
 
 if sys.version_info[0] < 3:
     from io import BytesIO as StringIO
@@ -17,15 +20,13 @@ try:
 except ImportError:
     from Queue import Queue
 
-# Tests depend on PyMongo's BSON implementation, but MockupDB itself does not.
 from bson import (Binary, Code, DBRef, Decimal128, MaxKey, MinKey, ObjectId,
                   Regex, SON, Timestamp)
 from bson.codec_options import CodecOptions
 from pymongo import MongoClient, message, WriteConcern
 
-from mockupdb import (_bson as mockup_bson, go, going,
-                      Command, CommandBase, Matcher, MockupDB, Request,
-                      OpInsert, OpMsg, OpQuery, QUERY_FLAGS)
+from mockupdb import (go, going, Command, CommandBase, Matcher, MockupDB,
+                      Request, OpInsert, OpMsg, OpQuery, QUERY_FLAGS)
 from tests import unittest  # unittest2 on Python 2.6.
 
 
@@ -209,17 +210,17 @@ class TestMatcher(unittest.TestCase):
         _id = '5a918f9fa08bff9c7688d3e1'
 
         for a, b in [
-            (Binary(b'foo'), mockup_bson.Binary(b'foo')),
-            (Code('foo'), mockup_bson.Code('foo')),
-            (Code('foo', {'x': 1}), mockup_bson.Code('foo', {'x': 1})),
-            (DBRef('coll', 1), mockup_bson.DBRef('coll', 1)),
-            (DBRef('coll', 1, 'db'), mockup_bson.DBRef('coll', 1, 'db')),
-            (Decimal128('1'), mockup_bson.Decimal128('1')),
-            (MaxKey(), mockup_bson.MaxKey()),
-            (MinKey(), mockup_bson.MinKey()),
-            (ObjectId(_id), mockup_bson.ObjectId(_id)),
-            (Regex('foo', 'i'), mockup_bson.Regex('foo', 'i')),
-            (Timestamp(1, 2), mockup_bson.Timestamp(1, 2)),
+            (Binary(b'foo'), Binary(b'foo')),
+            (Code('foo'), Code('foo')),
+            (Code('foo', {'x': 1}), Code('foo', {'x': 1})),
+            (DBRef('coll', 1), DBRef('coll', 1)),
+            (DBRef('coll', 1, 'db'), DBRef('coll', 1, 'db')),
+            (Decimal128('1'), Decimal128('1')),
+            (MaxKey(), MaxKey()),
+            (MinKey(), MinKey()),
+            (ObjectId(_id), ObjectId(_id)),
+            (Regex('foo', 'i'), Regex('foo', 'i')),
+            (Timestamp(1, 2), Timestamp(1, 2)),
         ]:
             # Basic case.
             self.assertTrue(
@@ -240,6 +241,18 @@ class TestMatcher(unittest.TestCase):
                 Matcher(Command('x', y=b)).matches(Command('x', y=a)),
                 "PyMongo %r != MockupDB %r" % (a, b))
 
+    def test_datetime(self):
+        server = MockupDB(auto_ismaster=True)
+        server.run()
+        client = MongoClient(server.uri)
+        # Python datetimes have microsecond precision, BSON only millisecond.
+        # Ensure this datetime matches itself despite the truncation.
+        dt = datetime.datetime(2018, 12, 1, 6, 6, 6, 12345)
+        doc = SON([('_id', 1), ('dt', dt)])
+        with going(client.db.collection.insert_one, doc):
+            server.receives(
+                OpMsg('insert', 'collection', documents=[doc])).ok()
+
 
 class TestAutoresponds(unittest.TestCase):
     def test_auto_dequeue(self):
@@ -318,6 +331,28 @@ class TestMockupDB(unittest.TestCase):
         self.assertEqual(ismaster['minWireVersion'], 1)
         self.assertEqual(ismaster['maxWireVersion'], 42)
 
+    @unittest.skipIf(sys.platform == 'win32', 'Windows')
+    def test_unix_domain_socket(self):
+        tmp = tempfile.NamedTemporaryFile(delete=False, suffix='.sock')
+        tmp.close()
+        server = MockupDB(auto_ismaster={'maxWireVersion': 3},
+                          uds_path=tmp.name)
+        server.run()
+        self.assertTrue(server.uri.endswith('.sock'),
+                        'Expected URI "%s" to end with ".sock"' % (server.uri,))
+        self.assertEqual(server.host, tmp.name)
+        self.assertEqual(server.port, 0)
+        self.assertEqual(server.address, (tmp.name, 0))
+        self.assertEqual(server.address_string, tmp.name)
+        client = MongoClient(server.uri)
+        with going(client.test.command, {'foo': 1}) as future:
+            server.receives().ok()
+
+        response = future()
+        self.assertEqual(1, response['ok'])
+        server.stop()
+        self.assertFalse(os.path.exists(tmp.name))
+
 
 class TestResponse(unittest.TestCase):
     def test_ok(self):
